import pytest
import subprocess
import shutil
import pathlib
import sys

from deprecation import fail_if_not_removed

from .utils import site_packages_readonly


data_files_combinations = [
    (
        # data file source
        ("share/test.txt",),
        # data file spec
        ("jupyter-packaging-test", "share", "**/*"),
        # data file target
        "jupyter-packaging-test/test.txt",
    ),
    (
        # data file source
        ("share/test.txt",),
        # data file spec
        ("jupyter-packaging-test/level1", "share", "**/[a-z]est.txt"),
        # data file target
        "jupyter-packaging-test/level1/test.txt",
    ),
    (
        # data file source
        ("level1/test/test.txt",),
        # data file spec
        ("jupyter-packaging-test", "level1/test", "**/*"),
        # data file target
        "jupyter-packaging-test/test.txt",
    ),
    (
        # data file source
        ("level1/test/test.txt",),
        # data file spec
        ("jupyter-packaging-test/level1/level2", "level1/test", "**/*"),
        # data file target
        "jupyter-packaging-test//level1/level2/test.txt",
    ),
    (
        # data file source
        ("level1/level2/test/test.txt",),
        # data file spec
        ("jupyter-packaging-test", "level1", "**/*"),
        # data file target
        "jupyter-packaging-test/level2/test/test.txt",
    ),
]


@pytest.mark.skipif(site_packages_readonly, reason="Site Packages are Read-only")
@fail_if_not_removed
@pytest.mark.parametrize("source,spec,target", data_files_combinations)
def test_develop(make_package_deprecated, source, spec, target):
    name = "jupyter_packaging_test_foo"
    package_dir = make_package_deprecated(
        name=name, data_files=source, data_files_spec=[spec]
    )
    target_path = pathlib.Path(sys.prefix).joinpath(target)
    if target_path.exists():
        shutil.rmtree(str(target_path.parent))
    subprocess.check_output(
        [sys.executable, "-m", "pip", "install", "-e", "."], cwd=str(package_dir)
    )
    assert target_path.exists()
    subprocess.check_output(
        [sys.executable, "-m", "pip", "uninstall", "-y", name], cwd=str(package_dir)
    )
    # This is something to fix later. uninstalling a package installed
    # with -e should be removed.
    assert target_path.exists()
    shutil.rmtree(str(target_path.parent))


@pytest.mark.skipif(site_packages_readonly, reason="Site Packages are Read-only")
@pytest.mark.parametrize("source,spec,target", data_files_combinations)
def test_install(make_package, source, spec, target):
    name = "jupyter_packaging_test_foo"
    package_dir = make_package(name=name, data_files=source, data_files_spec=[spec])
    target_path = pathlib.Path(sys.prefix).joinpath(target)
    if target_path.exists():
        shutil.rmtree(str(target_path.parent))
    subprocess.check_output(
        [sys.executable, "-m", "pip", "install", "."], cwd=str(package_dir)
    )
    assert target_path.exists()
    subprocess.check_output(
        [sys.executable, "-m", "pip", "uninstall", "-y", name], cwd=str(package_dir)
    )
    assert not target_path.exists()



import subprocess
from pathlib import Path
import sys
import sysconfig

from .utils import site_packages_readonly

import pytest


@pytest.mark.skipif(site_packages_readonly, reason="Site Packages are Read-only")
def test_install(make_package, tmp_path):
    name = "jupyter_packaging_test_foo"
    ensured_targets = [f"{name}/main.py"]
    package_dir = make_package(name=name, ensured_targets=ensured_targets)
    subprocess.check_output(
        [sys.executable, "-m", "pip", "install", "."], cwd=str(package_dir)
    )
    # Get site packages where the package is installed.
    sitepkg = Path(sysconfig.get_paths()["purelib"])
    installed_file = sitepkg / f"{name}/main.py"
    assert installed_file.exists()
    excluded_file = sitepkg / f"{name}/exclude.py"
    assert not excluded_file.exists()
    subprocess.check_output(
        [sys.executable, "-m", "pip", "uninstall", name, "-y"], cwd=str(package_dir)
    )
    assert not installed_file.exists()


@pytest.mark.skipif(site_packages_readonly, reason="Site Packages are Read-only")
def test_install_hybrid(make_hybrid_package, tmp_path):
    name = "jupyter_packaging_test_foo"
    ensured_targets = [f"{name}/main.py", f"{name}/generated.js"]
    package_dir = make_hybrid_package(
        name=name,
        ensured_targets=ensured_targets,
        skip_if_exists=[f"{name}/generated.js"],
    )
    subprocess.check_output(
        [sys.executable, "-m", "pip", "install", "."], cwd=str(package_dir)
    )
    # Get site packages where the package is installed.
    sitepkg = Path(sysconfig.get_paths()["purelib"])
    installed_py_file = sitepkg / f"{name}/main.py"
    installed_js_file = sitepkg / f"{name}/generated.js"
    assert installed_py_file.exists()
    assert installed_js_file.exists()
    content = installed_js_file.read_text(encoding="utf-8")
    Path(f"{package_dir}/{name}/generated.js").write_text(content, encoding="utf-8")
    excluded_file = sitepkg / f"{name}/exclude.py"
    assert not excluded_file.exists()
    subprocess.check_output(
        [sys.executable, "-m", "pip", "install", "."], cwd=str(package_dir)
    )
    assert content == installed_js_file.read_text(encoding="utf-8")
    subprocess.check_output(
        [sys.executable, "-m", "pip", "uninstall", name, "-y"], cwd=str(package_dir)
    )
    assert not installed_py_file.exists()
    assert not installed_js_file.exists()


@pytest.mark.skipif(site_packages_readonly, reason="Site Packages are Read-only")
def test_install_missing(make_package, tmp_path):
    name = "jupyter_packaging_test_foo"
    ensured_targets = [f"{name}/missing.py"]
    package_dir = make_package(name=name, ensured_targets=ensured_targets)
    with pytest.raises(subprocess.CalledProcessError):
        subprocess.check_output(
            [sys.executable, "-m", "pip", "install", "."], cwd=str(package_dir)
        )
    subprocess.check_output(
        [sys.executable, "-m", "pip", "install", "-e", "."], cwd=str(package_dir)
    )



from pathlib import Path
from unittest.mock import call

from setuptools.dist import Distribution

from jupyter_packaging.setupbase import npm_builder, wrap_installers


def test_wrap_installers():
    called = False

    def func():
        nonlocal called
        called = True

    cmd_class = wrap_installers(
        pre_dist=func, pre_develop=func, post_dist=func, post_develop=func
    )

    for name in ["pre_dist", "pre_develop", "post_dist", "post_develop"]:
        cmd_class[name](Distribution()).run()
        assert called
        called = False


def test_npm_builder(mocker):
    which = mocker.patch("jupyter_packaging.setupbase.which")
    run = mocker.patch("jupyter_packaging.setupbase.run")
    builder = npm_builder()
    which.return_value = ["foo"]
    builder()
    cwd = Path.cwd()
    run.assert_has_calls(
        [call(["npm", "install"], cwd=cwd), call(["npm", "run", "build"], cwd=cwd)]
    )


def test_npm_build_skip(mocker):
    which = mocker.patch("jupyter_packaging.setupbase.which")
    run = mocker.patch("jupyter_packaging.setupbase.run")
    mocker.patch("jupyter_packaging.setupbase.skip_npm", True)
    builder = npm_builder()
    which.return_value = ["foo"]
    builder()
    run.assert_not_called()


def test_npm_builder_yarn(tmp_path, mocker):
    which = mocker.patch("jupyter_packaging.setupbase.which")
    run = mocker.patch("jupyter_packaging.setupbase.run")
    tmp_path.joinpath("yarn.lock").write_text("hello")
    builder = npm_builder(path=tmp_path)
    which.return_value = ["foo"]
    builder()
    run.assert_has_calls(
        [
            call(["yarn", "install"], cwd=tmp_path),
            call(["yarn", "run", "build"], cwd=tmp_path),
        ]
    )


def test_npm_builder_missing_yarn(tmp_path, mocker):
    which = mocker.patch("jupyter_packaging.setupbase.which")
    run = mocker.patch("jupyter_packaging.setupbase.run")
    tmp_path.joinpath("yarn.lock").write_text("hello")
    builder = npm_builder(path=tmp_path)
    which.side_effect = ["", "foo"]
    builder()
    run.assert_has_calls(
        [
            call(["npm", "install"], cwd=tmp_path),
            call(["npm", "run", "build"], cwd=tmp_path),
        ]
    )


def test_npm_builder_not_stale(tmp_path, mocker):
    which = mocker.patch("jupyter_packaging.setupbase.which")
    run = mocker.patch("jupyter_packaging.setupbase.run")
    is_stale = mocker.patch("jupyter_packaging.setupbase.is_stale")
    is_stale.return_value = False
    builder = npm_builder(build_dir=tmp_path, source_dir=tmp_path)
    which.return_value = ["foo"]
    builder()
    run.assert_not_called()


def test_npm_builder_no_npm(mocker):
    which = mocker.patch("jupyter_packaging.setupbase.which")
    run = mocker.patch("jupyter_packaging.setupbase.run")
    is_stale = mocker.patch("jupyter_packaging.setupbase.is_stale")
    is_stale.return_value = False
    builder = npm_builder()
    which.return_value = []
    builder()
    run.assert_not_called()



from jupyter_packaging.setupbase import is_stale


def test_destination_is_not_stale(source_dir, destination_dir):
    assert is_stale(str(destination_dir), str(source_dir)) is False


def test_root_file_causes_stale(source_dir, destination_dir):
    source_dir.join("file1.txt").setmtime(30000)
    assert is_stale(str(destination_dir), str(source_dir)) is True


def test_sub_file_causes_stale(source_dir, destination_dir):
    source_dir.join("sub", "subfile2.txt").setmtime(30000)
    assert is_stale(str(destination_dir), str(source_dir)) is True


def test_folder_mtime_does_not_prevent_stale(source_dir, destination_dir):
    source_dir.join("sub", "subfile2.txt").setmtime(30000)
    destination_dir.setmtime(40000)
    destination_dir.join("sub").setmtime(40000)
    destination_dir.setmtime(40000)
    assert is_stale(str(destination_dir), str(source_dir)) is True


def test_folder_mtime_does_not_cause_stale(source_dir, destination_dir):
    source_dir.setmtime(40000)
    source_dir.join("sub").setmtime(40000)
    source_dir.setmtime(40000)
    assert is_stale(str(destination_dir), str(source_dir)) is False


# This behavior might not always be wanted?
# The alternative is to check whether ALL files in destination is newer
# than the newest file in source (more conservative).
def test_only_newest_files_determine_stale(source_dir, destination_dir):
    source_dir.join("file1.txt").setmtime(30000)
    destination_dir.join("file1.rtf").setmtime(40000)
    assert is_stale(str(destination_dir), str(source_dir)) is False


def test_unstale_on_equal(source_dir):
    assert is_stale(str(source_dir), str(source_dir)) is False


def test_file_vs_dir(source_dir, destination_dir):
    assert is_stale(str(destination_dir.join("file1.rtf")), str(source_dir)) is False
    source_dir.join("file2.txt").setmtime(30000)
    assert is_stale(str(destination_dir.join("file1.rtf")), str(source_dir)) is True


def test_dir_vs_file(source_dir, destination_dir):
    assert is_stale(str(destination_dir), str(source_dir.join("file1.txt"))) is False
    source_dir.join("file1.txt").setmtime(30000)
    assert is_stale(str(destination_dir), str(source_dir.join("file1.txt"))) is True


def test_file_vs_file(source_dir, destination_dir):
    assert (
        is_stale(
            str(destination_dir.join("file1.rtf")), str(source_dir.join("file1.txt"))
        )
        is False
    )
    source_dir.join("file1.txt").setmtime(30000)
    assert (
        is_stale(
            str(destination_dir.join("file1.rtf")), str(source_dir.join("file1.txt"))
        )
        is True
    )


def test_empty_dir(source_dir, tmpdir):
    empty_dir = tmpdir.mkdir("empty")
    assert is_stale(str(empty_dir), str(source_dir)) is True
    assert is_stale(str(source_dir), str(empty_dir)) is False
    assert is_stale(str(empty_dir), str(empty_dir)) is False



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import importlib
from pathlib import Path
import sys

from deprecation import deprecated
from setuptools.build_meta import (
    get_requires_for_build_wheel,
    get_requires_for_build_sdist,
    prepare_metadata_for_build_wheel,
    build_sdist as orig_build_sdist,
    build_wheel as orig_build_wheel,
)
import tomlkit

from jupyter_packaging.setupbase import __version__


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """Build a wheel with an optional pre-build step."""
    builder = _get_build_func()
    if builder:
        builder()
    val = orig_build_wheel(
        wheel_directory,
        config_settings=config_settings,
        metadata_directory=metadata_directory,
    )
    _ensure_targets()
    return val


def build_sdist(sdist_directory, config_settings=None):
    """Build an sdist with an optional pre-build step."""
    builder = _get_build_func()
    if builder:
        builder()
    val = orig_build_sdist(sdist_directory, config_settings=config_settings)
    _ensure_targets()
    return val


@deprecated(
    deprecated_in="0.8",
    removed_in="1.0",
    current_version=__version__,
    details="Use `factory =` instead",
)
def _handle_deprecated_metadata():
    pass


def _get_build_func():
    pyproject = Path("pyproject.toml")
    if not pyproject.exists():
        return
    data = tomlkit.loads(pyproject.read_text(encoding="utf-8"))
    if "tool" not in data:
        return
    if "jupyter-packaging" not in data["tool"]:
        return
    if "builder" not in data["tool"]["jupyter-packaging"]:
        return
    section = data["tool"]["jupyter-packaging"]

    # Handle deprecated "func" builder kwarg
    if "func" in section["builder"]:
        _handle_deprecated_metadata()
        if not "factory" in section["builder"]:
            section["builder"]["factory"] = section["builder"]["func"]

    if "factory" not in section["builder"]:
        raise ValueError("Missing `factory` specifier for builder")

    factory_data = section["builder"]["factory"]
    mod_name, _, factory_name = factory_data.rpartition(".")

    if "options" in section and "skip-if-exists" in section["options"]:
        skip_if_exists = section["options"]["skip-if-exists"]
        if all(Path(path).exists() for path in skip_if_exists):
            return None

    # If the module fails to import, try importing as a local script
    try:
        mod = importlib.import_module(mod_name)
    except ImportError:
        try:
            sys.path.insert(0, str(Path.cwd()))
            mod = importlib.import_module(mod_name)
        finally:
            sys.path.pop(0)

    factory = getattr(mod, factory_name)
    kwargs = section.get("build-args", {})
    return factory(**kwargs)


def _ensure_targets():
    pyproject = Path("pyproject.toml")
    if not pyproject.exists():
        return
    data = tomlkit.loads(pyproject.read_text(encoding="utf-8"))
    if "tool" not in data:
        return
    if "jupyter-packaging" not in data["tool"]:
        return
    section = data["tool"]["jupyter-packaging"]
    if "options" in section and "ensured-targets" in section["options"]:
        targets = section["options"]["ensured-targets"]
        missing = [t for t in targets if not Path(t).exists()]
        if missing:
            raise ValueError(("missing files: %s" % missing))



import os
import sys
from subprocess import check_call
from unittest.mock import patch, call

import pytest

from jupyter_packaging.build_api import build_wheel, build_sdist


TOML_CONTENT = """
[tool.jupyter-packaging.builder]
factory = "foo.main"

[tool.jupyter-packaging.build-args]
fizz = "buzz"
"""

FOO_CONTENT = r"""
from pathlib import Path
def main(fizz=None):
    Path('foo.txt').write_text(f'fizz={fizz}', encoding='utf-8')
"""

BAD_CONTENT = """
[tool.jupyter-packaging.builder]
bar = "foo.main"
"""

ENSURED_CONTENT = """
[tool.jupyter-packaging.options]
ensured-targets = ["foo.txt"]
"""

SKIP_IF_EXISTS = """
[tool.jupyter-packaging.options]
skip-if-exists = ["foo.txt"]
"""


def test_build_wheel_no_toml(tmp_path):
    os.chdir(tmp_path)
    orig_wheel = patch("jupyter_packaging.build_api.orig_build_wheel")
    build_wheel(tmp_path)
    orig_wheel.assert_called_with(
        tmp_path, config_settings=None, metadata_directory=None
    )


def test_build_wheel(tmp_path, mocker):
    os.chdir(tmp_path)
    tmp_path.joinpath("foo.py").write_text(FOO_CONTENT)
    tmp_path.joinpath("pyproject.toml").write_text(
        TOML_CONTENT + ENSURED_CONTENT, encoding="utf-8"
    )
    orig_wheel = mocker.patch("jupyter_packaging.build_api.orig_build_wheel")
    build_wheel(tmp_path)
    orig_wheel.assert_called_with(
        tmp_path, config_settings=None, metadata_directory=None
    )
    data = tmp_path.joinpath("foo.txt").read_text(encoding="utf-8")
    assert data == "fizz=buzz"

    content = TOML_CONTENT.replace("buzz", "fizz") + SKIP_IF_EXISTS
    tmp_path.joinpath("pyproject.toml").write_text(content, encoding="utf-8")
    build_wheel(tmp_path)
    data = tmp_path.joinpath("foo.txt").read_text(encoding="utf-8")
    assert data == "fizz=buzz"


def test_build_wheel_bad_toml(tmp_path, mocker):
    os.chdir(tmp_path)
    tmp_path.joinpath("foo.py").write_text(FOO_CONTENT)
    tmp_path.joinpath("pyproject.toml").write_text(BAD_CONTENT, encoding="utf-8")
    orig_wheel = mocker.patch("jupyter_packaging.build_api.orig_build_wheel")
    with pytest.raises(ValueError):
        build_wheel(tmp_path)
    orig_wheel.assert_not_called()


def test_build_wheel_no_toml(tmp_path, mocker):
    os.chdir(tmp_path)
    orig_wheel = mocker.patch("jupyter_packaging.build_api.orig_build_wheel")
    build_wheel(tmp_path)
    orig_wheel.assert_called_with(
        tmp_path, config_settings=None, metadata_directory=None
    )


def test_build_sdist(tmp_path, mocker):
    os.chdir(tmp_path)
    tmp_path.joinpath("foo.py").write_text(FOO_CONTENT)
    tmp_path.joinpath("pyproject.toml").write_text(
        TOML_CONTENT + ENSURED_CONTENT, encoding="utf-8"
    )
    orig_sdist = mocker.patch("jupyter_packaging.build_api.orig_build_sdist")
    build_sdist(tmp_path)
    orig_sdist.assert_called_with(tmp_path, config_settings=None)
    data = tmp_path.joinpath("foo.txt").read_text(encoding="utf-8")
    assert data == "fizz=buzz"


def test_build_sdist_bad_toml(tmp_path, mocker):
    os.chdir(tmp_path)
    tmp_path.joinpath("foo.py").write_text(FOO_CONTENT)
    tmp_path.joinpath("pyproject.toml").write_text(BAD_CONTENT, encoding="utf-8")
    orig_sdist = mocker.patch("jupyter_packaging.build_api.orig_build_sdist")
    with pytest.raises(ValueError):
        build_sdist(tmp_path)
    orig_sdist.assert_not_called()


def test_build_sdist_no_toml(tmp_path, mocker):
    os.chdir(tmp_path)
    orig_sdist = mocker.patch("jupyter_packaging.build_api.orig_build_sdist")
    build_sdist(tmp_path)
    orig_sdist.assert_called_with(tmp_path, config_settings=None)


def test_build_package(make_package):
    package_dir = make_package()
    pyproject = package_dir / "pyproject.toml"
    text = pyproject.read_text(encoding="utf-8")
    text = text.replace("setuptools.build_meta", "jupyter_packaging.build_api")
    text += TOML_CONTENT
    pyproject.write_text(text, encoding="utf-8")
    package_dir.joinpath("foo.py").write_text(FOO_CONTENT, encoding="utf-8")
    check_call([sys.executable, "-m", "build"], cwd=package_dir)
    data = package_dir.joinpath("foo.txt").read_text(encoding="utf-8")
    assert data == "fizz=buzz"


def test_deprecated_metadata(make_package):
    package_dir = make_package()
    pyproject = package_dir / "pyproject.toml"
    text = pyproject.read_text(encoding="utf-8")
    text = text.replace("setuptools.build_meta", "jupyter_packaging.build_api")
    text += TOML_CONTENT
    text = text.replace("factory =", "func =")
    pyproject.write_text(text, encoding="utf-8")
    package_dir.joinpath("foo.py").write_text(FOO_CONTENT, encoding="utf-8")
    check_call([sys.executable, "-m", "build"], cwd=package_dir)
    data = package_dir.joinpath("foo.txt").read_text(encoding="utf-8")
    assert data == "fizz=buzz"
